from __future__ import annotations

import os

from collections import defaultdict
from typing import (
    Any, ClassVar, Dict, List,
)

import param  # type: ignore
import yaml

from .base import MultiTypeComponent
from .config import config


class Auth(MultiTypeComponent):
    """
    An AuthPlugin is given the auth specfication and can apply arbitrary
    transforms to it.
    """

    auth_type: ClassVar[str | None] = None

    @classmethod
    def _import_module(cls, component_type: str):
        pass

    def transform(self, spec: Dict[str, Any]) -> Dict[str, Any]:
        return dict(spec)


class YamlAuthMapperPlugin(Auth):
    """
    The YamlAuthMapperPlugin uses a Yaml file to map auth keys
    allowing more concise declarations for individual dashboards, e.g.
    the following yaml specification:

        group:
          admins:
            email:
              - lumen@holoviz.org
              - panel@holoviz.org
              - param@holoviz.org
          users:
            email:
              - philipp@holoviz.org

    will expand this auth spec:

        auth:
          group:
            - admins

    to this expanded spec:

        auth:
          email:
            - lumen@holoviz.org
            - panel@holoviz.org
            - param@holoviz.org
            - philipp@holoviz.org
    """

    yaml_file = param.Filename()

    auth_type: ClassVar[str] = 'yaml'

    @classmethod
    def _validate_yaml_file(cls, yaml_file, spec, context):
        if config.root:
            yaml_file = os.path.join(config.root, yaml_file)
        return yaml_file

    def transform(self, spec: Dict[str, Any]) -> Dict[str, Any]:
        spec = dict(spec)
        with open(self.yaml_file) as f:
            text = f.read()
        mapping = yaml.load(text, Loader=yaml.Loader)
        new: Dict[str, List[str]] = defaultdict(list)
        for origin, replacements in mapping.items():
            values = spec.pop(origin)
            for val in values:
                for key, new_vals in replacements.get(val, {}).items():
                    new[key] += new_vals
        spec.update(new)
        return spec



from __future__ import annotations

from typing import ClassVar

import param  # type: ignore

from panel import panel
from panel.reactive import ReactiveHTML
from panel.widgets import FileDownload

from .util import catch_and_notify


class DownloadButton(ReactiveHTML):

    callback = param.Callable(precedence=-1)

    color = param.Color(default='grey', allow_None=True)

    data = param.String()

    filename = param.String()

    hide = param.Boolean(default=False)

    size = param.Integer(default=20)

    _template: ClassVar[str] = """
    <style>
    .download-button {
      position: absolute;
      top: 0px;
      right: 0px;
      width: {{ size }}px;
      height: {{ size }}px;
      z-index: 10000;
      opacity: {% if hide %}0{% else %}1{% endif %};
      transition-delay: 0.5s;
      transition: 0.5s;
      cursor: pointer;
      font-size: {{ size }}px;
      {% if color %}color: {{ color }};{% endif %}
    }

    .download-button:hover {
      transition: 0.5s;
      opacity: 1;
    }

    .download-button:focus {
      opacity: 1;
    }
    </style>
    <span id="download-button" onclick="${_on_click}" class="download-button">
      <i class="fas fa-download"></i>
    </span>
    """

    _scripts = {
        'data': """
          if (data.data == null || !data.data.length)
            return
          const byteString = atob(data.data.split(',')[1]);

          // separate out the mime component
          const mimeString = data.data.split(',')[0].split(':')[1].split(';')[0];

          // Reset data

          data.data = '';

          // write the bytes of the string to an ArrayBuffer
          const ab = new ArrayBuffer(byteString.length);
          const ia = new Uint8Array(ab);
          for (let i = 0; i < byteString.length; i++) {
            ia[i] = byteString.charCodeAt(i);
          }

          // write the ArrayBuffer to a blob, and you're done
          var bb = new Blob([ab], { type: mimeString });

          var link = document.createElement('a');

          link.href = URL.createObjectURL(bb)
          link.download = data.filename
          link.click()
        """
    }

    def __init__(self, object=None, **params):
        params['sizing_mode'] = 'stretch_width'
        if object is not None:
            object = panel(object)
            params['object'] = object
        super().__init__(**params)

    @catch_and_notify("Download failed")
    def _on_click(self, event=None):
        file_input = FileDownload(callback=self.callback, filename=self.filename)
        file_input._transfer()
        self.data = file_input.data


class IconButton(ReactiveHTML):

    disabled = param.Boolean(default=False)

    color = param.Color(default=None)

    icon = param.String(default=None, doc="""
      The FontAwesome icon to use.""")

    size = param.Integer(default=12, bounds=(0, None))

    _template = """
      <i id="icon-button" class="fas ${icon}" style="font-size: ${size}px; color: ${color}" onclick=${script('clicked')}></i>
    """

    _scripts = {
        'clicked': """
          if (data.disabled)
            return
          data.disabled = true;
          view._send_event('button', 'click', state.event)
        """,
        'disabled': """
          icon_button.style.cursor = data.disabled ? "not-allowed": "inherit";
        """
    }

    _event = 'dom_event'

    def __init__(self, **params):
        super().__init__(**params)
        self._callbacks = []
        self._disabled_watcher = None

    def _enable_button(self, event):
        if self._disabled_watcher:
            self.param.unwatch(self._disabled_watcher)
        self.disabled = False

    @param.depends('size', watch=True, on_init=True)
    def _update_height(self):
        self.height = self.size

    def on_click(self, callback):
        self._callbacks.append(callback)

    def js_on_click(self, args={}, code=""):
        from panel.links import Callback
        return Callback(self, code={'event:'+self._event: code}, args=args)

    def _button_click(self, event=None):
        for cb in self._callbacks:
            cb(event)
        self._disabled_watcher = self.param.watch(self._enable_button, ['disabled'])



import contextlib
import os
import tempfile

from unittest.mock import Mock

import pandas as pd
import panel as pn
import pytest

from bokeh.document import Document  # type: ignore
from hvplot.tests.util import makeMixedDataFrame

from lumen.config import config
from lumen.sources.base import FileSource, Source
from lumen.state import state
from lumen.variables.base import Variables

with contextlib.suppress(Exception):
    # From Dask 2024.3.0 they now use `dask_expr` by default
    # https://github.com/dask/dask/issues/10995
    import dask

    dask.config.set({"dataframe.query-planning": False})


@pytest.fixture
def set_root():
    root = config._root
    def _set_root(root):
        config._root = root
    yield _set_root
    config._root = root

@pytest.fixture
def make_filesource():
    root = config._root
    def create(root, **kwargs):
        config._root = root
        source = FileSource(tables={'test': 'test.csv'},  kwargs={'parse_dates': ['D']}, **kwargs)
        state.sources['original'] = source
        return source
    yield create
    config._root = root
    for source in state.global_sources.values():
        source.clear_cache()
    state.global_sources.clear()

@pytest.fixture
def make_variable_filesource():
    root = config._root
    def create(root, **kwargs):
        config._root = root
        state._variable = Variables.from_spec({'tables': {'type': 'constant', 'default': {'test': 'test.csv'}}})
        source = Source.from_spec({
            'type': 'file',
            'tables': '$variables.tables',
            'kwargs': {'parse_dates': ['D']}
        })
        state.sources['original'] = source
        return source
    yield create
    config._root = root
    for source in state.global_sources.values():
        source.clear_cache()
    state.global_sources.clear()
    state._variable = None

@pytest.fixture
def mixed_df():
    pytest.importorskip("pyarrow", "7.0", "Pyarrow is not out on Python 3.12 yet")
    string = pd.get_option('mode.string_storage')
    pd.set_option('mode.string_storage', 'pyarrow')
    df = makeMixedDataFrame()
    df['C'] = df.C.astype("string")
    yield df
    pd.set_option('mode.string_storage', string)

@pytest.fixture
def mixed_df_object_type():
    string = pd.get_option('mode.string_storage')
    pd.set_option('mode.string_storage', 'python')
    df = makeMixedDataFrame()
    yield df
    pd.set_option('mode.string_storage', string)

@pytest.fixture
def yaml_file():
    tf = tempfile.NamedTemporaryFile(mode='w', suffix='.yaml', delete=False)
    yield tf
    tf.close()
    os.unlink(tf.name)

@pytest.fixture
def state_userinfo():
    mock_state = Mock()
    mock_state.user_info = {'email': 'lumen@holoviz.org', 'user': 'lumen'}
    mock_state.user = 'lumen'
    state = pn.state
    pn.state = mock_state
    yield
    pn.state = state

@pytest.fixture(autouse=True)
def clear_state():
    "Clear the state after the execution of each test"
    yield
    state.global_sources.clear()
    state.global_filters.clear()
    state._configs.clear()
    state._config = None
    state._specs.clear()
    state._spec.clear()
    state._loadings.clear()
    state._loading = None
    state._sources.clear()
    state._source.clear()
    state._filters.clear()
    state._filter.clear()
    state._pipelines.clear()
    state._pipeline.clear()
    state._variables.clear()
    state._variable = None

@pytest.fixture
def document():
    doc = Document()
    with pn.io.server.set_curdoc(doc):
        yield

@pytest.fixture
def cachedir():
    tmp_dir = tempfile.TemporaryDirectory()
    yield tmp_dir.name
    tmp_dir.cleanup()


@pytest.fixture
def expected_filtered_df(source_tables, table_column_value_type):
    table, column, value, type = table_column_value_type
    df = source_tables[table]

    if type == 'single_value':
        if value is None:
            df = df[df[column].isnull()]
        else:
            df = df[df[column] == value]

    elif type == 'range':
        begin, end = value
        df = df[(df[column] >= begin) & (df[column] <= end)]

    elif type == 'range_list':
        conditions = False
        for range in value:
            begin, end = range
            conditions |= ((df[column] >= begin) & (df[column] <= end))
        df = df[conditions]

    elif type == 'list':
        df = df[df[column].isin(value)]

    elif type == 'date':
        df = df[df[column] == pd.to_datetime(value)]

    elif type == 'date_range':
        begin, end = value
        df = df[(df[column] >= pd.to_datetime(begin)) & (df[column] <= pd.to_datetime(end))]

    return df

@pytest.fixture
def penguins_file(tmp_path):
    # created with
    # df = pd.read_csv(url).sample(5).reset_index(drop=True).to_csv()
    fn = tmp_path / "penguins.csv"
    penguins = """
    ,species,island,bill_length_mm,bill_depth_mm,flipper_length_mm,body_mass_g,sex,year
    0,Gentoo,Biscoe,46.7,15.3,219.0,5200.0,male,2007
    1,Adelie,Dream,42.2,18.5,180.0,3550.0,female,2007
    2,Gentoo,Biscoe,51.1,16.3,220.0,6000.0,male,2008
    3,Adelie,Biscoe,45.6,20.3,191.0,4600.0,male,2009
    4,Adelie,Torgersen,37.8,17.1,186.0,3300.0,,2007
    """
    fn.write_text(penguins)
    return fn



from __future__ import annotations

import param  # type: ignore

from .intake_sql import IntakeBaseSQLSource


class IntakeBaseDremioSource(IntakeBaseSQLSource):
    """
    Base class with common parameters for Dremio sources.
    """

    cert = param.String(default="Path to certificate file")

    dask = param.Boolean(default=False, doc="""
        Whether to return a dask DataFrame.""")

    uri = param.String(doc="URI of the Dremio server.")

    tls = param.Boolean(default=False, doc="Enable encryption")

    username = param.String(default=None, doc="Dremio username")

    password = param.String(default=None, doc="Dremio password or token")

    def _get_source(self, table):
        # Fuzzy matching to ignore quoting issues
        normalized_table = table.replace('"', '').lower()
        tables = self.get_tables()
        normalized_tables = [t.replace('"', '').lower() for t in tables]
        if table not in tables and normalized_table in normalized_tables:
            table = tables[normalized_tables.index(normalized_table)]
        return super()._get_source(table)


class IntakeDremioSource(IntakeBaseDremioSource):
    """
    `IntakeDremioSource` allows querying Dremio catalog tables and views.

    When provided with the `uri` of the Dremio server and credentials
    to authenticate with the Dremio instance all available tables can
    be queried via this `Source`.

    Requires the intake-dremio package to be installed.
    """

    source_type = 'intake_dremio'

    def __init__(self, **params):
        from intake_dremio.dremio_cat import DremioCatalog  # type: ignore
        super().__init__(**params)
        self.cat = DremioCatalog(
            self.uri, cert=self.cert, tls=self.tls, username=self.username,
            password=self.password
        )


class IntakeDremioSQLSource(IntakeBaseDremioSource):
    """
    `IntakeDremioSQLSource` allows querying a subset of
    Dremio catalog tables and views via custom SQL expressions.

    When provided with the `uri` of the Dremio server and credentials
    to authenticate with the Dremio instance, unlike `IntakeDremioSource`,
    only the tables specified in the `tables` parameter can be used.

    Requires the intake-dremio package to be installed.
    """

    tables = param.Dict(default={}, doc="""
        Mapping of table names to desired SQL expressions""")

    source_type = 'intake_dremio_sql'

    def __init__(self, **params):
        from intake_dremio.intake_dremio import DremioSource
        super().__init__(**params)
        self.cat = {
            table: DremioSource(
                sql_expr=sql_expr,
                cert=self.cert,
                uri=self.uri,
                tls=self.tls,
                username=self.username,
                password=self.password,
            )
            for table, sql_expr in self.tables.items()
        }



import datetime as dt
import os

import pandas as pd
import pytest
import yaml

from lumen.sources.intake import IntakeSource


@pytest.fixture
def source():
    root = os.path.dirname(__file__)
    return IntakeSource(
        uri=os.path.join(root, 'catalog.yml'), root=root
    )


@pytest.fixture
def source_tables(mixed_df):
    df_test = mixed_df.copy()
    df_test_sql = mixed_df.copy()
    df_test_sql_none = mixed_df.copy()
    df_test_sql_none['C'] = ['foo1', None, 'foo3', None, 'foo5']
    tables = {
        'test': df_test,
        'test_sql': df_test_sql,
        'test_sql_with_none': df_test_sql_none,
    }
    return tables


def test_intake_resolve_module_type():
    assert IntakeSource._get_type('lumen.sources.intake_sql.IntakeSource') is IntakeSource
    assert IntakeSource.source_type == 'intake'


def test_intake_source_from_file(source, mixed_df):
    pd.testing.assert_frame_equal(source.get('test'), mixed_df)


def test_intake_source_from_dict(mixed_df):
    root = os.path.dirname(__file__)
    with open(os.path.join(root, 'catalog.yml')) as f:
        catalog = yaml.load(f, Loader=yaml.Loader)
    source = IntakeSource(catalog=catalog, root=root)
    pd.testing.assert_frame_equal(source.get('test'), mixed_df)


@pytest.mark.parametrize(
    "table_column_value_type", [
        ('test', 'A', 1, 'single_value'),
        ('test', 'A', (1, 3), 'range'),
        ('test', 'A', [(0, 1), (3, 4)], 'range_list'),
        ('test', 'C', 'foo2', 'single_value'),
        ('test', 'C', ['foo1', 'foo3'], 'list'),
        ('test', 'D', dt.datetime(2009, 1, 2), 'single_value'),
        ('test', 'D', (dt.datetime(2009, 1, 2), dt.datetime(2009, 1, 5)), 'range'),
        ('test', 'D', dt.date(2009, 1, 2), 'date'),
        ('test', 'D', (dt.date(2009, 1, 2), dt.date(2009, 1, 5)), 'date_range'),
    ]
)
@pytest.mark.parametrize("dask", [True, False])
def test_intake_filter(source, table_column_value_type, dask, expected_filtered_df):
    table, column, value, _ = table_column_value_type
    kwargs = {column: value}
    filtered = source.get(table, __dask=dask, **kwargs)
    pd.testing.assert_frame_equal(filtered, expected_filtered_df)



import pandas as pd
import panel as pn
import param  # type: ignore

from .util import resolve_module_reference


class JSONSchema(pn.pane.PaneBase):

    default_layout = param.ClassSelector(default=pn.Column, class_=(pn.layout.Panel),
                                         is_instance=False, doc="""
        Defines the layout the model(s) returned by the pane will
        be placed in.""")

    multi = param.Boolean(default=True, doc="""
        Whether to pick multi-select and range widgets by default.""")

    properties = param.List(default=[], doc="""
        If set this serves as a whitelist of properties to display on
        the supplied schema object.""")

    schema = param.Dict(doc="""
        A valid jsonschema""")

    widgets = param.Dict(doc="""
        Dictionary of widget overrides, mapping from parameter name
        to widget class.""")

    _precedence = ['enum', 'type']

    _unpack = True

    _select_widget = pn.widgets.Select
    _multi_select_widget = pn.widgets.MultiSelect
    _bounded_number_widget = pn.widgets.FloatSlider
    _bounded_number_range_widget = pn.widgets.RangeSlider
    _unbounded_number_widget = pn.widgets.FloatInput
    _list_select_widget = pn.widgets.MultiSelect
    _bounded_int_widget = pn.widgets.IntSlider
    _bounded_int_range_widget = pn.widgets.IntRangeSlider
    _unbounded_int_widget = pn.widgets.IntInput
    _string_widget = pn.widgets.TextInput
    _boolean_widget = pn.widgets.Checkbox
    _literal_widget = pn.widgets.LiteralInput
    _unbounded_date_widget = pn.widgets.DatePicker
    _date_range_widget = pn.widgets.DateRangeSlider

    # Not available until Panel 0.11
    try:
        _datetime_range_widget = pn.widgets.DatetimeRangeInput
    except Exception:
        _datetime_range_widget = pn.widgets.DateRangeSlider

    # Not available until Panel 0.12
    try:
        _unbounded_datetime_widget = pn.widgets.DatetimePicker
        _datetime_range_widget = pn.widgets.DatetimeRangePicker
    except Exception:
        _unbounded_datetime_widget = pn.widgets.DatetimeInput

    def _array_type(self, schema):
        if 'items' in schema and not schema.get('additionalItems', True):
            items = schema['items']
            if len(items) == 2:
                if all(s['type'] == 'number' for s in items):
                    return pn.widgets.RangeSlider, {}
                elif all(s['type'] == 'integer' for s in items):
                    return pn.widgets.IntRangeSlider, {}
            elif 'enum' in items:
                return self._list_select_type, {'options': items['enum']}
        else:
            return self._literal_widget, {'type': (list, tuple)}

    def _number_type(self, schema):
        if 'inclusiveMaximum' in schema and 'inclusiveMinimum' in schema:
            if self.multi:
                wtype = self._bounded_number_range_widget
            else:
                wtype = self._bounded_number_widget
            return wtype, {'start': float(schema['inclusiveMinimum']),
                           'end': float(schema['inclusiveMaximum'])}
        else:
            return self._unbounded_number_widget, {}

    def _integer_type(self, schema):
        if 'inclusiveMaximum' in schema and 'inclusiveMinimum' in schema:
            if self.multi:
                wtype = self._bounded_int_range_widget
            else:
                wtype = self._bounded_int_widget
            return wtype, {'start': int(schema['inclusiveMinimum']),
                           'end': int(schema['inclusiveMaximum'])}
        else:
            return self._unbounded_int_widget, {'step': 1}

    def _boolean_type(self, schema):
        return self._boolean_widget, {}

    def _string_type(self, schema):
        if schema.get('format') in ('date', 'date-time', 'datetime'):
            if 'formatMaximum' in schema and 'formatMinimum' in schema:
                dt_min = pd.to_datetime(schema['formatMinimum'])
                if dt_min.tz:
                    dt_min = dt_min.astimezone('UTC').tz_localize(None)
                dt_max = pd.to_datetime(schema['formatMaximum'])
                if dt_max.tz:
                    dt_max = dt_max.astimezone('UTC').tz_localize(None)
                dt_range = {'start': dt_min.to_pydatetime(),
                            'end': dt_max.to_pydatetime()}
                if schema.get('format') == 'date':
                    wtype = self._date_range_widget
                else:
                    wtype = self._datetime_range_widget
                return wtype, dt_range
            if 'inclusiveMaximum' in schema and 'inclusiveMinimum' in schema:
                dt_min = pd.to_datetime(schema['inclusiveMinimum'])
                if dt_min.tz:
                    dt_min = dt_min.astimezone('UTC').tz_localize(None)
                dt_max = pd.to_datetime(schema['inclusiveMaximum'])
                if dt_max.tz:
                    dt_max = dt_max.astimezone('UTC').tz_localize(None)
                dt_range = {'start': dt_min.to_pydatetime(),
                            'end': dt_max.to_pydatetime()}
                if schema.get('format') == 'date':
                    wtype = self._date_range_widget
                else:
                    wtype = self._datetime_range_widget
                return wtype, dt_range
            elif schema.get('format') == 'date':
                return self._unbounded_date_widget, {}
            else:
                return self._unbounded_datetime_widget, {}
        return self._string_widget, {}

    def _enum(self, schema):
        wtype = self._multi_select_widget if self.multi else self._select_widget
        try:
            options = sorted(schema['enum'])
        except TypeError:
            options = schema['enum']
        return wtype, {'options': options}

    def __init__(self, object=None, schema=None, **params):
        if schema is not None:
            params['schema'] = schema
        super().__init__(object, **params)
        self._widgets = {}
        self._update_widgets_from_schema()

    @param.depends('schema', watch=True)
    def _update_widgets_from_schema(self):
        if not self.schema:
            self.layout[:] = []
            return

        values = {} if self.object is None else self.object
        widgets = []
        for p, schema in self.schema.items():
            if self.properties and p not in self.properties:
                continue
            for prop in self._precedence:
                if prop in schema:
                    break

            if self.widgets is None or p not in self.widgets:
                wtype, kwargs = self._widget_type(prop, schema)
            elif isinstance(self.widgets[p], dict):
                kwargs = dict(self.widgets[p])
                if 'type' in kwargs:
                    wtype = kwargs.pop('type')
                else:
                    wtype = self._widget_type(prop, schema)
            else:
                _, kwargs = self._widget_type(prop, schema)
                wtype = self.widgets[p]

            if isinstance(wtype, str):
                wtype = resolve_module_reference(wtype, pn.widgets.Widget)

            if isinstance(wtype, pn.widgets.Widget):
                widget = wtype
            else:
                if p in values:
                    kwargs['value'] = values[p]
                widget = wtype(name=schema.get('title', p), **kwargs)
            self._widgets[p] = widget
            widgets.append(widget)
        self.layout[:] = widgets

    def _widget_type(self, prop, schema):
        if prop == 'enum':
            wtype, kwargs = self._enum(schema)
        else:
            wtype, kwargs = getattr(self, '_%s_%s' % (schema[prop], prop))(schema)
        return wtype, kwargs

    @param.depends('object', watch=True)
    def _update_widget_values(self):
        if not self.object:
            return
        for p, value in self.object.items():
            self._widgets[p].value = value

    def _get_model(self, doc, root=None, parent=None, comm=None):
        model = self.layout._get_model(doc, root, parent, comm)
        self._models[root.ref['id']] = (model, parent)
        return model

    def _cleanup(self, root):
        self.layout._cleanup(root)
        super(JSONSchema, self)._cleanup(root)



from __future__ import annotations

import html
import importlib
import importlib.util
import os
import traceback

from concurrent.futures import Future, ThreadPoolExecutor
from typing import (
    TYPE_CHECKING, Any, Callable, ClassVar, Dict, List, Tuple, Type,
)

import panel as pn
import param  # type: ignore
import yaml

from panel.io.resources import CSS_URLS
from panel.template.base import BasicTemplate
from panel.viewable import Viewable, Viewer
from typing_extensions import Literal

from .auth import Auth
from .base import Component, MultiTypeComponent
from .config import (
    _DEFAULT_LAYOUT, _LAYOUTS, _TEMPLATES, _THEMES, Template, config,
)
from .filters.base import ConstantFilter, Filter, WidgetFilter  # noqa
from .layout import Layout
from .panel import IconButton
from .pipeline import Pipeline
from .sources.base import RESTSource, Source  # noqa
from .state import state
from .transforms.base import Transform  # noqa
from .util import catch_and_notify, expand_spec, resolve_module_reference
from .validation import (
    ValidationError, match_suggestion_message, validate_callback,
)
from .variables.base import Variable, Variables
from .views.base import Download, View  # noqa

if TYPE_CHECKING:
    from bokeh.server.contexts import BokehSessionContext


def load_yaml(yaml_spec: str, **kwargs) -> Dict[str, Any]:
    expanded = expand_spec(yaml_spec, config.template_vars, **kwargs)
    return yaml.load(expanded, Loader=yaml.Loader)


class Config(Component):
    """
    `Config` provides high-level configuration options for the
    :class:`lumen.dashboard.Dashboard`.
    """

    auto_update = param.Boolean(default=None, constant=True, doc="""
        Whether changes in filters, transforms and references automatically
        trigger updates in the data or whether an update has to be triggered
        manually using the update event or the update button in the UI."""
    )

    background_load = param.Boolean(default=False, constant=True, doc="""
        Whether to load any layouts in the background.""")

    editable = param.Boolean(default=False, constant=True, doc="""
        Whether the dashboard specification is editable from within
        the deployed dashboard.""")

    layout = param.Selector(default=_DEFAULT_LAYOUT, constant=True, objects=_LAYOUTS,
                            check_on_set=False, doc="""
        Overall layout of the dashboard.""")

    loading_spinner = param.Selector(default='arc', constant=True, objects=[
        'arc', 'arcs', 'bar', 'dots', 'petal'], doc="""
        Loading indicator to use when component loading parameter is set.""")

    loading_color = param.Color(default='#0072B5', constant=True, doc="""
        Color of the loading indicator.""")

    logo = param.String(default=None, constant=True, doc="""
        A logo to add to the theme.""")

    ncols = param.Integer(default=3, bounds=(1, None), constant=True, doc="""
        Number of columns to lay out layouts in.""")

    on_session_created = param.Callable(constant=True, doc="""
        Callback that fires when a user session is created.""")

    on_session_destroyed = param.Callable(constant=True, doc="""
        Callback that fires when a user session is destroyed.""")

    on_loaded = param.Callable(constant=True, doc="""
        Callback that fires when a user frontend session is fully loaded.""")

    on_error = param.Callable(constant=True, doc="""
        Callback that fires if an error occurs in a dashboard callback.
        The exception is passed as the first argument.""")

    on_update = param.Callable(constant=True, doc="""
        Callback that fires when a pipeline is updated. The updated
        pipeline is passed as the first argument.""")

    reloadable = param.Boolean(default=True, constant=True, doc="""
        Whether to allow reloading data from source(s) using a button.""")

    sync_with_url = param.Boolean(default=False, constant=True, doc="""
        Whether to sync current state of the application.""")

    title = param.String(default="Lumen Dashboard", constant=True, doc="""
        The title of the dashboard.""")

    template = Template(default=_TEMPLATES['material'], constant=True, doc="""
        The Panel template to render the dashboard into.""")

    theme = param.Selector(default=_THEMES['default'], objects=_THEMES, constant=True,
                           check_on_set=False, doc="""
        The Panel template theme to style the dashboard with.""")

    show_traceback = param.Boolean(default=True, doc="""
        Whether to show the traceback if an error happens.""")

    _valid_keys: ClassVar[Literal['params']] = 'params'
    _validate_params: ClassVar[bool] = True

    @classmethod
    def _extract_template_type(cls, template: str | Dict[str, Any]) -> Tuple[str, Dict[str, Any]]:
        if isinstance(template, dict):
            template_type = template.get("type", _TEMPLATES["material"])
            template_params = template.copy()
        else:
            template_type = template
            template_params = {"type": template}
        return template_type, template_params

    @classmethod
    def _serialize_template(cls, template_params: Dict[str, Any]) -> str | Dict[str, Any]:
        if len(template_params) == 1:
            return template_params["type"]
        else:
            return template_params

    @classmethod
    def _validate_layout(cls, layout, spec, context):
        if layout not in _LAYOUTS:
            msg = f'Config layout {layout!r} could not be found. Layout must be one of {list(_LAYOUTS)}.'
            raise ValidationError(msg, spec, 'layout')
        return layout

    @classmethod
    def _validate_template(
        cls, template: str | Dict[str, Any], spec: Dict[str, Any], context: Dict[str, Any]
    ) -> Dict[str, Any]:
        template, template_params = cls._extract_template_type(template)
        if template in _TEMPLATES:
            return cls._serialize_template(template_params)
        elif '.' not in template:
            raise ValidationError(
                f'Config template {template!r} not found. Template must be one '
                f'of {list(_TEMPLATES)} or an absolute import path.', spec, template
            )
        *paths, name = template.split('.')
        path = '.'.join(paths)
        try:
            module = importlib.import_module(path)
        except Exception as e:
            raise ValidationError(
                f'Config template {path!r} module could not be imported '
                f'errored with: {e}.', spec, path
            )
        if not hasattr(module, name):
            raise ValidationError(
                f'Config template {name!r} was not found in {path!r} module.',
                spec, name
            )
        template_cls = getattr(module, name)
        if not issubclass(template_cls, BasicTemplate):
            raise ValidationError(
                f'Config template \'{path}.{name}\' is not a valid Panel template.',
                spec, 'template'
            )
        template_params["type"] = template_cls
        return cls._serialize_template(template_params)

    @classmethod
    def _validate_callback(cls, callback: Callable[..., Any] | str) -> Callable[Any, None]:
        if isinstance(callback, str):
            return resolve_module_reference(callback)
        return callback

    @classmethod
    def _validate_on_error(
        cls, on_error: Callable[[Type[Exception]], None] | str,
        spec: Dict[str, Any], context: Dict[str, Any]
    ) -> str:
        cb = cls._validate_callback(on_error)
        validate_callback(cb, ('exception',), what='on_error callback')
        return cb

    @classmethod
    def _validate_on_loaded(
        cls, on_loaded: Callable[[], None] | str, spec: Dict[str, Any],
        context: Dict[str, Any]
    ) -> str:
        cb = cls._validate_callback(on_loaded)
        validate_callback(cb, (), what='on_loaded callback')
        return cb

    @classmethod
    def _validate_on_update(
        cls, on_update: Callable[[Pipeline], None] | str, spec: Dict[str, Any],
        context: Dict[str, Any]
    ) -> str:
        cb = cls._validate_callback(on_update)
        validate_callback(cb, ('pipeline',), what='on_update callback')
        return cb

    @classmethod
    def _validate_on_session_created(
        cls, on_session_created: Callable[[], None] | str, spec: Dict[str, Any],
        context: Dict[str, Any]
    ) -> str:
        cb = cls._validate_callback(on_session_created)
        validate_callback(cb, (), what='on_session_created callback')
        return cb

    @classmethod
    def _validate_on_session_destroyed(
        cls, on_session_destroyed: Callable[[BokehSessionContext], None] | str,
        spec: Dict[str, Any], context: Dict[str, Any]
    ) -> str:
        cb = cls._validate_callback(on_session_destroyed)
        validate_callback(cb, ('session_context',), what='on_session_destroyed callback')
        return cb

    @classmethod
    def _validate_theme(
        cls, theme: str, spec: Dict[str, Any], context: Dict[str, Any]
    ) -> str:
        if theme not in _THEMES:
            msg = f'Config theme {theme!r} could not be found. Theme must be one of {list(_THEMES)}.'
            raise ValidationError(msg, spec, 'theme')
        return theme

    @classmethod
    def from_spec(cls, spec: Dict[str, Any] | str) -> 'Dashboard':
        if isinstance(spec, str):
            raise ValueError(
                "Config cannot be materialized by reference. Please pass "
                "full specification for the Config."
            )
        if 'template' in spec:
            template, template_params = cls._extract_template_type(spec['template'])
            if template in _TEMPLATES:
                template_cls = _TEMPLATES[template]
            else:
                template_cls = resolve_module_reference(template, BasicTemplate)
            template_params['type'] = template_cls
            spec['template'] = cls._serialize_template(template_params)

        if 'theme' in spec:
            spec['theme'] = _THEMES[spec['theme']]
        if 'layout' in spec:
            spec['layout'] = _LAYOUTS[spec['layout']]
        for key in list(spec):
            if key.startswith('on_'):
                spec[key] = getattr(cls, f'_validate_{key}')(spec[key], spec, {})

        return cls(**spec)

    def to_spec(self, context: Dict[str, Any] | None = None) -> Dict[str, Any]:
        spec = super().to_spec(context=context)
        if 'layout' in spec:
            spec['layout'] = {v: k for k, v in _LAYOUTS.items()}[spec['layout']]
        if 'template' in spec:
            template = spec['template']
            template, template_params = self._extract_template_type(template)
            template_params['type'] = {v: k for k, v in _TEMPLATES.items()}.get(
                template, f'{template.__module__}.{template.__name__}')
            spec['template'] = self._serialize_template(template_params)
        if 'theme' in spec:
            spec['theme'] = {v: k for k, v in _THEMES.items()}[spec['theme']]
        for key in list(spec):
            if key.startswith('on_'):
                cb = spec[key]
                ref = f'{cb.__module__}.{cb.__name__}'
                if ref.startswith('__main__'):
                    raise ValueError(
                        f"Cannot serialize 'config.{key}' callback because "
                        "it was defined in the __main__ scope. Please move the "
                        "callback into an importable module."
                    )
                spec[key] = ref
        return spec

    def __init__(self, **params):
        super().__init__(**params)
        pn.config.param.update(
            loading_spinner=self.loading_spinner,
            loading_color=self.loading_color
        )
        if self.on_session_destroyed:
            try:
                pn.state.on_session_destroyed(self.on_session_destroyed)
            except RuntimeError:
                pass

    def construct_template(self):
        template, template_params = self._extract_template_type(self.template)
        template_params.update({'title': self.title, 'theme': self.theme})
        if self.logo:
            template_params['logo'] = self.config.logo
        template_params.pop('type')
        return template(**template_params)


class Defaults(Component):
    """
    `Defaults` to apply to the component classes.
    """

    download = param.Dict(default={}, doc="Defaults for the Download object")

    filters = param.List(doc="Defaults for Filter objects.", item_type=dict)

    sources = param.List(doc="Defaults for Source objects.", item_type=dict)

    transforms = param.List(doc="Defaults for Transform objects.", item_type=dict)

    views = param.List(doc="Defaults for View objects.", item_type=dict)

    _valid_keys: ClassVar[Literal['params']] = 'params'
    _validate_params: ClassVar[bool] = True

    @classmethod
    def _validate_defaults(
        cls, defaults_type: Type[MultiTypeComponent], defaults: List[Dict[str, Any]], spec: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        highlight = f'{defaults_type.__name__.lower()}s'
        if not isinstance(defaults, list):
            raise ValidationError(
                f'Defaults must be defined as a list not as {type(defaults)}.',
                spec, highlight
            )
        for default in defaults:
            if 'type' not in default:
                raise ValidationError(
                    'Defaults must declare the type the defaults apply to.',
                    spec, highlight
                )
            default = dict(default)
            dtype = default.pop('type')
            obj_type = defaults_type._get_type(dtype)
            for p in default:
                if p in obj_type.param:
                    pobj = obj_type.param[p]
                    try:
                        pobj._validate(default[p])
                    except Exception as e:
                        msg = f"The default for {obj_type.__name__} {p!r} parameter failed validation: {str(e)}"
                        raise ValidationError(msg, default, p)
                else:
                    msg = (
                        f'Default for {obj_type.__name__} {p!r} parameter cannot be set '
                        'as there is no such parameter.'
                    )
                    msg = match_suggestion_message(p, list(obj_type.param), msg)
                    raise ValidationError(msg, default, p)
        return defaults

    @classmethod
    def _validate_download(
        cls, download_defaults: Dict[str, Any], spec: Dict[str, Any], context: Dict[str, Any]
    ):
        if not isinstance(download_defaults, dict):
            msg = f'Defaults for Download component must be declared as a dictionary, not as a {type(download_defaults)}.'
            raise ValidationError(msg, spec, 'download:')
        for p in download_defaults:
            if p in Download.param:
                pobj = Download.param[p]
                try:
                    pobj._validate(download_defaults[p])
                except Exception as e:
                    msg = f"The default for Download {p!r} parameter failed validation: {str(e)}"
                    raise ValidationError(msg, download_defaults, p)
                continue
            msg = (
                f'Default for Download {p!r} parameter cannot be set as there '
                'is no such parameter.'
            )
            msg = match_suggestion_message(p, list(Download.param), msg)
            raise ValidationError(msg, download_defaults, p)
        return download_defaults

    @classmethod
    def _validate_filters(
        cls, filter_defaults: List[Dict[str, Any]], spec: Dict[str, Any], context: Dict[str, Any]
    ):
        return cls._validate_defaults(Filter, filter_defaults, spec)

    @classmethod
    def _validate_sources(
        cls, source_defaults: List[Dict[str, Any]], spec: Dict[str, Any], context: Dict[str, Any]
    ):
        return cls._validate_defaults(Source, source_defaults, spec)

    @classmethod
    def _validate_transforms(
        cls, transform_defaults: List[Dict[str, Any]], spec: Dict[str, Any], context: Dict[str, Any]
    ):
        return cls._validate_defaults(Transform, transform_defaults, spec)

    @classmethod
    def _validate_views(
        cls, view_defaults: List[Dict[str, Any]], spec: Dict[str, Any], context: Dict[str, Any]
    ):
        return cls._validate_defaults(View, view_defaults, spec)

    def apply(self):
        Download.param.update(self.download)
        for obj, defaults in ((Filter, self.filters), (Source, self.sources),
                              (Transform, self.transforms), (View, self.views)):
            for default in defaults:
                params = dict(default)
                obj_type = obj._get_type(params.pop('type'))
                obj_type.param.set_param(**params)


class AuthSpec(Component):
    """
    `AuthSpec` allows specifying a spec that is validated against panel.state.user_info.

    To enable `AuthSpec` you must configure an OAuth provider when deploying
    an application with `panel.serve`.
    """

    case_sensitive = param.Boolean(default=False, constant=True, doc="""
        Whether auth validation is case-sensitive or not.""")

    spec = param.Dict({}, constant=True, doc="""
        Dictionary of keys and values to match the pn.state.user_info
        against.""")

    _allows_refs: ClassVar[bool] = False

    @classmethod
    def from_spec(cls, spec: Dict[str, Any] | str) -> 'AuthSpec':
        if isinstance(spec, str):
            raise ValueError(
                "AuthSpec cannot be materialized by reference. Please pass "
                "full specification for AuthSpec."
            )
        spec = dict(spec)
        case_sensitive = spec.pop('case_sensitive', cls.case_sensitive)
        plugins = spec.pop('plugins', [])
        for plugin_spec in plugins:
            plugin = Auth.from_spec(plugin_spec)
            spec = plugin.transform(spec)
        return cls(spec=spec, case_sensitive=case_sensitive)

    @property
    def authorized(self) -> bool:
        if not self.spec:
            return True
        elif pn.state.user_info is None:
            return config.dev
        authorized = True
        for k, value in self.spec.items():
            if not isinstance(value, list): value = [value]
            if not self.case_sensitive:
                value = [v.lower() if isinstance(v, str) else v for v in value]
            value = [v.strip() if isinstance(v, str) else v for v in value]
            if k in pn.state.user_info:
                user_value = pn.state.user_info[k]
                if not isinstance(user_value, list):
                    user_value = [user_value]
                if not self.case_sensitive:
                    user_value = [
                        uv.lower() if isinstance(uv, str) else uv
                        for uv in user_value
                    ]
                authorized &= any(uv == v for v in value for uv in user_value)
        return authorized

    @classmethod
    def validate(
        cls, spec: Dict[str, Any] | str, context: Dict[str, Any] | None = None
    ) -> Dict[str, Any] | str:
        """
        Validates the component specification given the validation context.

        Arguments
        -----------
        spec: dict | str
          The specification for the component being validated (or a referene to the component)
        context: dict
          Validation context contains the specification of all previously validated components,
          e.g. to allow resolving of references.

        Returns
        --------
        Validated specification.
        """
        spec = super().validate(spec, context)
        plugins = [
            Auth.validate(plugin, context) for plugin in spec.get('plugins', [])
        ]
        if plugins:
            spec['plugins'] = plugins
        return spec


class Dashboard(Component, Viewer):

    auth = param.ClassSelector(default=AuthSpec(), class_=AuthSpec, doc="""
        AuthSpec object which validates the auth spec against pn.state.user_info.""")

    config = param.ClassSelector(default=Config(), class_=Config, doc="""
        High-level configuration options for the dashboard.""")

    defaults = param.ClassSelector(default=Defaults(), class_=Defaults, doc="""
        Defaults to apply to component classes.""")

    layouts = param.List(default=[], item_type=Layout, doc="""
        List of layouts monitoring some source.""")

    # Specification configuration
    _allows_refs = False

    _valid_keys = [
        'variables', 'auth', 'defaults', 'config', 'sources',
        'pipelines', 'layouts'
    ]

    def __init__(self, specification=None, **params):
        self._load_global = params.pop('load_global', True)
        for subobj in (Auth, Config, Defaults):
            pname = subobj.__name__.lower()
            if pname in params and isinstance(params[pname], dict):
                params[pname] = subobj.from_spec(params[pname])
        if specification is None:
            specification = state.to_spec(**params)
            load_vars = False
        else:
            load_vars = True
        if isinstance(specification, dict):
            state.spec = self.validate(specification)
            self._yaml = yaml.dump(specification)
            self._yaml_file = 'local'
            root = params.pop('root', os.getcwd())
        else:
            self._yaml_file = specification
            root = os.path.abspath(os.path.dirname(self._yaml_file))
        self._root = config.root = root
        self._edited = False
        self._debug = params.pop('debug', False)
        self._rerender_watchers = {}
        super().__init__(**params)
        self._init_config()

        # Initialize from spec
        config.load_local_modules()
        if isinstance(specification, dict):
            state.resolve_views()
        else:
            self._load_specification(from_file=True)
        self._apps = {}

        # Set up background loading
        self._thread_pool = None
        self._rendered = []

        # Initialize high-level settings
        self.auth = AuthSpec.from_spec(state.spec.get('auth', {}))
        self.config = Config.from_spec(state.spec.get('config', {}))
        self.defaults = Defaults.from_spec(state.spec.get('defaults', {}))
        if load_vars:
            self.variables = Variables.from_spec(state.spec.get('variables', {}))
        else:
            self.variables = state.variables
        self.defaults.apply()

        # Load and populate template
        self._template = self.config.construct_template()
        self._create_modal()
        self._create_main()
        self._create_sidebar()
        self._create_header()
        self._populate_template()

        state.config = self.config
        if pn.state._is_pyodide:
            self._render_dashboard()
            if self.config.on_loaded:
                pn.state.execute(self.config.on_loaded)
        else:
            pn.state.onload(self._render_dashboard)
            if self.config.on_loaded:
                pn.state.onload(self.config.on_loaded)
        if self.config.on_session_created:
            pn.state.execute(self.config.on_session_created)

    def _init_config(self):
        pn.config.notifications = True
        if CSS_URLS['font-awesome'] not in pn.config.css_files:
            pn.config.css_files.append(CSS_URLS['font-awesome'])

    def __panel__(self) -> Viewable:
        return self.layout()

    ##################################################################
    # Load specification
    ##################################################################

    def _render_dashboard(self):
        try:
            self._materialize_specification()
            self._render()
            self._main[:] = [self._layout]
        except Exception as e:
            self.param.warning(f'Rendering dashboard raised following error:\n\n {type(e).__name__}: {e}')
            self._main.loading = False
            if self.config.show_traceback:
                tb = html.escape(traceback.format_exc())
                alert = pn.pane.Alert(
                    f'<b>{type(e).__name__}</b>: {e}</br><pre style="overflow-y: scroll">{tb}</pre>',
                    alert_type="danger", sizing_mode='stretch_width'
                )
            else:
                alert = pn.pane.Alert("""
                    <div style="display: flex; justify-content: center; align-items: center; min-height:300px">
                        <h2>Something went wrong</h2>
                    </div>""",
                    alert_type="danger", sizing_mode='stretch_width',
                )
            self._main[:] = [alert]
        if isinstance(self._layout, pn.Tabs) and self.config.sync_with_url:
            pn.state.location.sync(self._layout, {'active': 'layout'})

    def _load_specification(self, from_file: bool = False):
        if self._yaml_file == 'local':
            return
        kwargs = {}
        if from_file or self._yaml is None:
            with open(self._yaml_file) as f:
                self._yaml = f.read()
        elif self._edited:
            kwargs = {'getenv': False, 'getshell': False, 'getoauth': False}
        state.spec = self.validate(load_yaml(self._yaml, **kwargs))
        state.resolve_views()

    def _load_layout(self, layout_spec: Dict[str, Any]) -> Layout:
        layout_spec = dict(layout_spec)
        if 'reloadable' not in layout_spec:
            layout_spec['reloadable'] = self.config.reloadable
        if 'auto_update' not in layout_spec and self.config.auto_update is not None:
            layout_spec['auto_update'] = self.config.auto_update
        layout = Layout.from_spec(layout_spec)
        self._rerender_watchers[layout] = layout.param.watch(
            self._render_layouts, 'rerender'
        )
        if isinstance(self._layout, pn.Tabs):
            layout.show_title = False
        layout.start()
        return layout

    def _background_load(self, i: int, spec: Dict[str, Any]) -> Layout:
        self.layouts[i] = layout = self._load_layout(spec)
        return layout

    def _materialize_specification(self, force: bool = False):
        if force or self._load_global or not state.global_sources:
            state.load_global_sources(clear_cache=force)
        if force or self._load_global or not state.pipelines:
            kwargs = {}
            if self.config.auto_update is not None:
                kwargs['auto_update'] = self.config.auto_update
            state.load_pipelines(**kwargs)
        if not self.auth.authorized:
            self.layouts = []
            return

        # Clean up old layouts
        for layout in self.layouts:
            if isinstance(layout, Layout) and layout in self._rerender_watchers:
                layout.param.unwatch(self._rerender_watchers[layout])
                del self._rerender_watchers[layout]

        layouts = []
        layout_specs = state.spec.get('layouts', [])
        nlayouts = len(layout_specs)
        if isinstance(self._layout, pn.Tabs) and self.config.background_load:
            self._thread_pool = ThreadPoolExecutor(max_workers=nlayouts-1)
        for i, layout_spec in enumerate(layout_specs):
            if state.loading_msg:
                state.loading_msg.object = (
                    f"Loading layout {layout_spec['title']!r} ({i + 1}/{nlayouts})..."
            )
            if isinstance(self._layout, pn.Tabs) and i != self._layout.active:
                if self.config.background_load:
                    item = self._thread_pool.submit(self._background_load, i, layout_spec)
                else:
                    item = None
                self._rendered.append(False)
            else:
                item = self._load_layout(layout_spec)
                self._rendered.append(True)
            layouts.append(item)
        self.layouts[:] = layouts
        if self._thread_pool is not None:
            self._thread_pool.shutdown()

    ##################################################################
    # Create UI
    ##################################################################

    def _create_header(self):
        self._header = pn.Row()
        self._reload_button = IconButton(
            icon='fa-sync', size=18, align=('end', 'center')
        )
        self._reload_button.on_click(self._reload)
        menu_items = []
        for yml in config.yamls:
            with open(yml) as f:
                spec = load_yaml(f.read())
            endpoint = os.path.basename(yml).split('.')[0]
            title = spec.get('config', {}).get('title', endpoint)
            i = 1
            while title in self._apps:
                i += 1
                title = f'{title} {i}'
            menu_items.append(title)
            self._apps[title] = endpoint
        if len(config.yamls) > 1:
            self._reload_button.margin = (7, 0, 0, 0)
            self._menu_button = pn.widgets.MenuButton(
                name='Select dashboard', items=menu_items, width=200,
                align='center', margin=(0, 25, 0, 0)
            )
            self._menu_button.on_click(self._navigate)
            self._header.append(self._menu_button)
        if self.config.reloadable:
            self._header.append(self._reload_button)
        if self.config.editable:
            self._header.append(self._edit_button)
        if 'auth' in state.spec:
            logout = IconButton(
                icon='fa-sign-out-alt', size=18, align=('end', 'center'),
            )
            logout.js_on_click(code="""
            window.location.href = '/logout'
            """)
            self._header.extend([
                pn.layout.HSpacer(),
                f'<b><font size="4.5em">User: {pn.state.user}</font></b>',
                logout
            ])

    def _create_main(self):
        layout_kwargs = {'sizing_mode': 'stretch_width', 'min_height': 400}
        if self.config.layout is pn.GridBox:
            layout_kwargs['ncols'] = self.config.ncols
        self._layout = self.config.layout(**layout_kwargs)
        styles = {'text-align': 'center', 'font-size': '1.8em', 'font-weight': 'bold'}
        style_params = {'styles': styles}
        state.loading_msg = pn.pane.HTML(
            'Loading...', align='center', width=400, height=400, **style_params
        )

        self._loading = pn.Column(
            pn.layout.HSpacer(), state.loading_msg, pn.layout.HSpacer(),
            sizing_mode='stretch_both'
        )
        self._main = pn.Column(
            self._loading, loading=True, sizing_mode='stretch_width',
        )
        if isinstance(self._layout, pn.Tabs):
            self._layout.param.watch(self._activate_filters, 'active')

    def _create_modal(self):
        if not self.config.editable:
            self._modal = pn.Column()
            return
        self._editor = pn.widgets.CodeEditor(
            value=self._yaml, filename=self._yaml_file,
            sizing_mode='stretch_both', min_height=600,
            theme='monokai'
        )
        self._edit_button = pn.widgets.Button(
            name='✎', width=50, css_classes=['reload'], margin=0,
            align='center'
        )
        self._editor.param.watch(self._edit, 'value')
        self._edit_button.js_on_click(code="""
        var modal = document.getElementById("pn-Modal")
        modal.style.display = "block"
        """)
        self._modal = pn.Column(
            f'## Edit {os.path.basename(self._yaml_file)}',
            self._editor, sizing_mode='stretch_both'
        )

    def _create_sidebar(self):
        self._sidebar = pn.Accordion(margin=0, sizing_mode='stretch_width')
        self._update_button = pn.widgets.Button(name='Apply Update')
        self._update_button.on_click(self._update_pipelines)

    def _populate_template(self):
        if self.config.editable:
            self._template.modal[:] = [self._modal]
        self._template.main[:] = [self._main]
        self._template.header[:] = [self._header]
        self._template.sidebar[:] = [self._sidebar]

    ##################################################################
    # Rendering API
    ##################################################################

    @catch_and_notify
    def _activate_filters(self, event: param.parameterized.Event):
        layout = self.layouts[event.new]
        rendered = self._rendered[event.new]
        if not rendered:
            spec = state.spec['layouts'][event.new]
            self._set_loading(spec['title'])
            try:
                if isinstance(layout, Future):
                    layout = layout.result()
                else:
                    layout = self._load_layout(spec)
            except Exception as e:
                self._main.loading = False
                raise e
            self.layouts[event.new] = layout
            self._layout[event.new] = layout.panels
            self._rendered[event.new] = True
        self._render_filters()
        self._main.loading = False

    def _edit(self, event: param.parameterized.Event):
        self._yaml = event.new
        self._edited = True

    def _navigate(self, event: param.parameterized.Event):
        if pn.state.location is None:
            return
        app = self._apps[event.new]
        pn.state.location.pathname = f'/{app}'
        pn.state.location.reload = True

    def _set_loading(self, name: str = '', reloading: bool = True):
        state.loading_msg.object = f'<b>Reloading {name}...</b>'
        if reloading:
            self._main.loading = True
            return
        if isinstance(self._layout, pn.GridBox):
            items = [pn.pane.HTML(width=self._layout[i].width) # type: ignore
                     for i in range(self._layout.ncols) if i < len(self._layout)]
            index = int(min(self._layout.ncols, (len(self._layout)-1)) / 2)
            if items:
                items[index] = self._loading
            else:
                items = [self._loading]
        elif isinstance(self._layout, pn.Tabs):
            items = list(zip(self._layout._names, list(self._layout.objects))) # type: ignore
            tab_name = items[self._layout.active][0]
            if name and tab_name != name:
                return
            items[self._layout.active] = (tab_name, self._loading)
        else:
            items = [self._loading]
        self._layout[:] = items
        self._main.loading = True

    def _open_modal(self, event: param.parameterized.Event):
        self._template.open_modal()

    def _get_global_filters(self) -> Tuple[List[Filter] | None, pn.Column | None]:
        views = []
        filters = []
        for layout in self.layouts:
            if layout is None or isinstance(layout, Future):
                continue
            for pipeline in layout._pipelines.values():
                for filt in pipeline.filters:
                    if ((all(isinstance(layout, (type(None), Future)) or any(filt in p.filters for p in layout._pipelines.values())
                             for layout in self.layouts) and
                         filt.panel is not None) and filt.shared) and filt not in filters:
                        views.append(filt.panel)
                        filters.append(filt)
        if not views or len(self.layouts) == 1:
            return None, None
        if not self.config.auto_update:
            views.append(self._update_button)
        return filters, pn.Column(*views, name='Filters', sizing_mode='stretch_width')

    def _update_pipelines(self, event: param.parameterized.Event | None = None):
        self._update_button.loading = True
        try:
            for layout in self.layouts:
                if layout is None or isinstance(layout, Future):
                    continue
                for pipeline in layout._pipelines.values():
                    pipeline.param.trigger('update')
        finally:
            self._update_button.loading = False

    def _render_filters(self):
        self._global_filters, global_panel = self._get_global_filters()
        filters = [] if global_panel is None else [global_panel]
        global_refs = [ref.split('.')[1] for ref in state.global_refs]
        self._variable_panel = self.variables.panel(global_refs)
        apply_button = not self.config.auto_update or len(self.layouts) == 1
        if self._variable_panel is not None:
            filters.append(self._variable_panel)
        for i, layout in enumerate(self.layouts):
            if isinstance(self._layout, pn.Tabs) and i != self._layout.active:
                continue
            elif layout is None or isinstance(layout, Future):
                spec = state.spec['layouts'][i]
                panel = pn.Column(name=spec['title'])
            else:
                panel = layout.get_filter_panel(skip=self._global_filters, apply_button=apply_button)
            if panel is not None:
                filters.append(panel)
        self._sidebar[:] = filters

    def _render_layouts(self, event: param.parameterized.Event | None = None):
        if event is not None:
            self._set_loading(event.obj.title, reloading=event.name == 'rerender')
        items = []
        for layout, spec in zip(self.layouts, state.spec.get('layouts', [])):
            if layout is None or isinstance(layout, Future):
                panel = pn.layout.VSpacer(name=spec['title'])
            else:
                panel = layout.panels
            items.append(panel)
        self._layout[:] = items
        self._main.loading = False

    def _render_unauthorized(self):
        auth_keys = list(state.spec.get('auth'))
        if len(auth_keys) == 1:
            auth_keys = repr(auth_keys[0])
        else:
            auth_keys = [repr(k) for k in auth_keys]
        if pn.state.user_info is None:
            error = """
            ## Authorization Error

            Cannot verify permissioning since OAuth is not enabled.
            """
        else:
            error = f"""
            ## Unauthorized User

            {pn.state.user} is not authorized. Ensure {pn.state.user}
            is permissioned correctly on the {auth_keys} field(s) in
            OAuth user data.
            """
        alert = pn.pane.Alert(error, alert_type='danger')
        if isinstance(self._layout, pn.Tabs):
            alert = ('Authorization Denied', alert)
        self._layout[:] = [alert]

    def _render(self):
        if self.auth.authorized:
            self._render_filters()
            if isinstance(self._layout, pn.Tabs):
                count = 1 + bool(self._global_filters) + bool(self._variable_panel)
                active = list(range(count))
            else:
                active = list(range(len(self._sidebar)))
            self._sidebar.active = active
            self._render_layouts()
        else:
            self._render_unauthorized()
        self._main.loading = False

    def _reload(self, *events: param.parameterized.Event):
        self._load_specification()
        self._materialize_specification(force=True)
        self._render()

    ##################################################################
    # Validation API
    ##################################################################

    @classmethod
    def _validate_pipelines(
        cls, pipeline_specs: Dict[str, Any], spec: Dict[str, Any], context: Dict[str, Any]
    ) -> Dict[str, Any]:
        if 'pipelines' not in context:
            context['pipelines'] = {}
        return cls._validate_dict_subtypes('pipelines', Pipeline, pipeline_specs, spec, context, context['pipelines'])

    @classmethod
    def _validate_sources(
        cls, source_specs: Dict[str, Any], spec: Dict[str, Any], context: Dict[str, Any]
    ) -> Dict[str, Any]:
        if 'sources' not in context:
            context['sources'] = {}
        return cls._validate_dict_subtypes('sources', Source, source_specs, spec, context, context['sources'])

    @classmethod
    def _validate_layouts(
        cls, layout_specs: List[Dict[str, Any] | str], spec: Dict[str, Any], context: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        if 'layouts' not in context:
            context['layouts'] = []
        return cls._validate_list_subtypes('layouts', Layout, layout_specs, spec, context, context['layouts'])

    @classmethod
    def _validate_variables(
        cls, variable_specs: Dict[str, Any], spec: Dict[str, Any], context: Dict[str, Any]
    ) -> Dict[str, Any]:
        if 'variables' not in context:
            context['variables'] = {}
        return cls._validate_dict_subtypes('variables', Variable, variable_specs, spec, context, context['variables'])

    ##################################################################
    # Public API
    ##################################################################

    @classmethod
    def validate(
        cls, spec: Dict[str, Any] | str, context: Dict[str, Any] | None = None
    ) -> Dict[str, Any]:
        """
        Validates the component specification given the validation context.

        Arguments
        -----------
        spec: dict
          The specification for the component being validated.

        Returns
        --------
        Validated specification.
        """
        if isinstance(spec, str):
            raise ValueError(
                'Dashboard must be materialized with a full specification. '
                'Materializing a Dashboard from a string is not supported.'
            )
        if context is not None:
            raise ValueError(
                'Dashboard.validate is already the top-level context '
                'you may not pass a context argument.'
            )
        # Backward compatibility for old naming (targets -> layouts)
        if 'targets' in spec:
            spec['layouts'] = spec.pop('targets')

        cls._validate_keys_(spec)
        cls._validate_required_(spec)
        return cls._validate_spec_(spec)

    def layout(self):
        """
        Returns a layout of the dashboard contents.
        """
        spinner = pn.indicators.LoadingSpinner(width=30, height=30, align=('end', 'center'))
        pn.state.sync_busy(spinner)

        title_html = f'<div style="color: white; font-size: 2em;">{self.config.title}</div>'
        styles = {'background': '#0072b5'}
        header_params = dict(styles=styles)
        return pn.Column(
            pn.Row(
                pn.pane.HTML(title_html, margin=(20, 20), align='center'),
                pn.layout.HSpacer(),
                spinner,
                sizing_mode='stretch_width',
                **header_params
            ),
            pn.Row(
                pn.Row(self._sidebar, max_width=320),
                pn.Row(self._layout, margin=(5, 5, 5, 20)),
                sizing_mode='stretch_width'
            ),
            sizing_mode='stretch_both'
        )

    def to_spec(self, context: Dict[str, Any] | None = None) -> Dict[str, Any]:
        """
        Exports the full specification to reconstruct this component.

        Returns
        -------
        Declarative specification of this component.
        """
        if context is not None:
            raise ValueError(
                'Dashboard.to_spec is already the top-level context '
                'you may not pass a context argument.'
            )
        spec: Dict[str, Any] = {}
        spec['variables'] = {
            name: variable.to_spec(context=spec)
            for name, variable in state.variables._vars.items()
        }
        spec['sources'] = {
            name: source.to_spec(context=spec)
            for name, source in state.sources.items()
        }
        spec['pipelines'] = {
            name: pipeline.to_spec(context=spec)
            for name, pipeline in state.pipelines.items()
        }
        spec.update(super().to_spec(context=spec))
        spec['layouts'] = [
            tspec if layout is None else layout
            for layout, tspec in zip(spec['layouts'], state.spec['layouts'])
        ]
        return {k: v for k, v in spec.items() if v != {}}

    def show(self, **kwargs):
        """
        Opens the dashboard in a new window.
        """
        self._template.show(title=self.config.title, **kwargs)

    def servable(self):
        """
        Marks the dashboard for serving with `panel serve`.
        """
        self._template.servable(title=self.config.title)



from __future__ import annotations

from typing import Any, Dict

import param  # type: ignore

from ..transforms.base import Filter
from ..transforms.sql import (
    SQLDistinct, SQLFilter, SQLLimit, SQLMinMax,
)
from ..util import get_dataframe_schema
from ..validation import ValidationError
from .base import cached, cached_schema
from .intake import IntakeBaseSource, IntakeSource


class IntakeBaseSQLSource(IntakeBaseSource):

    cache_per_query = param.Boolean(default=True, doc="""
        Whether to query the whole dataset or individual queries.""")

    filter_in_sql = param.Boolean(default=True, doc="")

    # Declare this source supports SQL transforms
    _supports_sql = True

    __abstract = True

    def _apply_transforms(self, source, sql_transforms):
        if not sql_transforms:
            return source
        sql_expr = source._sql_expr
        for sql_transform in sql_transforms:
            sql_expr = sql_transform.apply(sql_expr)
        return type(source)(**dict(source._init_args, sql_expr=sql_expr))

    def _get_source(self, table):
        try:
            source = self.cat[table]
        except KeyError as e:
            raise ValidationError(
                f"'{table}' table could not be found in Intake catalog. "
                f"Available tables include: {list(self.cat)}."
            ) from e
        return source

    def get_sql_expr(self, table):
        return self._get_source(table)._sql_expr

    @cached
    def get(self, table, **query):
        '''
        Applies SQL Transforms, creating new temp catalog on the fly
        and querying the database.
        '''
        dask = query.pop('__dask', self.dask)
        sql_transforms = query.pop('sql_transforms', [])
        source = self._get_source(table)
        if not hasattr(source, '_sql_expr'):
            if sql_transforms:
                raise ValueError(
                    'SQLTransforms cannot be applied to non-SQL based Intake source.'
                )
            return super().get(table, **query)
        conditions = list(query.items())
        if self.filter_in_sql:
            sql_transforms = [SQLFilter(conditions=conditions)] + sql_transforms
        source = self._apply_transforms(source, sql_transforms)
        df = self._read(source)
        if not self.filter_in_sql:
            df = Filter.apply_to(df, conditions=conditions)
        return df if dask or not hasattr(df, 'compute') else df.compute()

    @cached_schema
    def get_schema(
        self, table: str | None = None, limit: int | None = None
    ) -> Dict[str, Dict[str, Any]] | Dict[str, Any]:
        if table is None:
            tables = self.get_tables()
        else:
            tables = [table]

        schemas = {}
        sql_limit = SQLLimit(limit=limit or 1)
        for entry in tables:
            if not self.load_schema:
                schemas[entry] = {}
                continue
            source = self._get_source(entry)
            if not hasattr(source, '_sql_expr'):
                schemas[entry] = super().get_schema(table)
                continue
            data = self._read(self._apply_transforms(source, [sql_limit]))
            schemas[entry] = schema = get_dataframe_schema(data)['items']['properties']
            if limit:
                continue

            enums, min_maxes = [], []
            for name, col_schema in schema.items():
                if 'enum' in col_schema:
                    enums.append(name)
                elif 'inclusiveMinimum' in col_schema:
                    min_maxes.append(name)

            # Calculate enum schemas
            for col in enums:
                distinct_transforms = [SQLDistinct(columns=[col])]
                distinct = self._read(
                    self._apply_transforms(source, distinct_transforms)
                )
                schema[col]['enum'] = distinct[col].to_list()

            if not min_maxes:
                continue

            # Calculate numeric schemas
            minmax_data = self._read(
                self._apply_transforms(source, [SQLMinMax(columns=min_maxes)])
            )
            for col in min_maxes:
                kind = data[col].dtype.kind
                if kind in 'iu':
                    cast = int
                elif kind == 'f':
                    cast = float
                else:
                    cast = lambda v: v
                schema[col]['inclusiveMinimum'] = cast(minmax_data[f'{col}_min'].iloc[0])
                schema[col]['inclusiveMaximum'] = cast(minmax_data[f'{col}_max'].iloc[0])
        return schemas if table is None else schemas[table]


class IntakeSQLSource(IntakeBaseSQLSource, IntakeSource):
    """
    `IntakeSQLSource` extends the `IntakeSource` with support for SQL data.

    In addition to the standard intake support for reading catalogs
    the `IntakeSQLSource` computes the schema by querying the database
    instead of loading all the data into memory and allows for
    `SQLTransform` to be applied when querying the SQL database.
    """

    source_type = 'intake_sql'



import os

import pandas as pd
import pytest

from lumen.sources.base import DerivedSource, Source


@pytest.fixture
def original(make_filesource):
    root = os.path.dirname(__file__)
    return make_filesource(root)


@pytest.fixture
def expected_table(mixed_df):
    return mixed_df.iloc[:3]


@pytest.fixture
def expected_schema():
    schema = {
        'A': {'inclusiveMaximum': 2.0, 'inclusiveMinimum': 0.0, 'type': 'number'},
        'B': {'inclusiveMaximum': 1.0, 'inclusiveMinimum': 0.0, 'type': 'number'},
        'C': {'enum': ['foo1', 'foo2', 'foo3'], 'type': 'string'},
        'D': {
            'format': 'datetime',
            'type': 'string',
            'inclusiveMaximum': '2009-01-05T00:00:00',
            'inclusiveMinimum': '2009-01-01T00:00:00'
        }
    }
    return schema


@pytest.fixture
def mirror_mode_spec():
    spec = {
        'type': 'derived',
        'source': 'original',
    }
    return spec


@pytest.fixture
def tables_mode_spec():
    spec = {
        'type': 'derived',
        'tables': {'derived': {'source': 'original', 'table': 'test'}}
    }
    return spec


def test_derived_source_resolve_module_type():
    assert DerivedSource._get_type('lumen.sources.base.DerivedSource') is DerivedSource


def test_derived_mirror_source(original, mirror_mode_spec):
    derived = Source.from_spec(mirror_mode_spec)
    assert derived.get_tables() == original.get_tables()
    assert derived.get_schema() == original.get_schema()
    for table in original.get_tables():
        pd.testing.assert_frame_equal(derived.get(table), original.get(table))


@pytest.mark.parametrize("additional_spec", [
    ('transforms', [{'type': 'iloc', 'end': 3}]),
    ('filters', [{'type': 'constant', 'field': 'A', 'value': (0, 2)}]),
])
def test_derived_mirror_source_apply(
        original, mirror_mode_spec, additional_spec, expected_table, expected_schema
):
    spec_key, spec_value = additional_spec
    mirror_mode_spec[spec_key] = spec_value
    derived = Source.from_spec(mirror_mode_spec)
    assert derived.get_tables() == original.get_tables()
    assert derived.get_schema('test') == expected_schema
    pd.testing.assert_frame_equal(derived.get('test'), expected_table)


def test_derived_tables_source(original, tables_mode_spec, mixed_df):
    derived = Source.from_spec(tables_mode_spec)
    assert derived.get_tables() == ['derived']
    pd.testing.assert_frame_equal(derived.get('derived'), mixed_df)
    assert original.get_schema('test') == derived.get_schema('derived')


@pytest.mark.parametrize("additional_spec", [
    ('transforms', [{'type': 'iloc', 'end': 3}]),
    ('filters', [{'type': 'constant', 'field': 'A', 'value': (0, 2)}]),
])
def test_derived_tables_source_apply(
        original, tables_mode_spec, additional_spec, expected_table, expected_schema
):
    spec_key, spec_value = additional_spec
    tables_mode_spec['tables']['derived'][spec_key] = spec_value
    derived = Source.from_spec(tables_mode_spec)
    assert derived.get_tables() == ['derived']
    pd.testing.assert_frame_equal(derived.get('derived'), expected_table)
    assert derived.get_schema('derived') == expected_schema


def test_derived_get_query_cache(original, mirror_mode_spec):
    derived = Source.from_spec(mirror_mode_spec)
    df = derived.get('test')
    cache_key = derived._get_key('test')
    assert cache_key in derived._cache
    cached_df = derived._cache[cache_key]
    pd.testing.assert_frame_equal(cached_df, df)


def test_derived_clear_cache(original, mirror_mode_spec):
    derived = Source.from_spec(mirror_mode_spec)
    derived.get('test')
    cache_key = derived._get_key('test')
    assert cache_key in derived._cache
    derived.clear_cache()
    assert len(derived._cache) == 0



import param as _param  # type: ignore

from .config import config  # noqa
from .dashboard import Dashboard  # noqa
from .filters.base import Filter  # noqa
from .layout import Layout  # noqa
from .pipeline import Pipeline  # noqa
from .sources.base import Source  # noqa
from .state import state  # noqa
from .transforms.base import Transform  # noqa
from .views.base import View  # noqa

__version__ = str(_param.version.Version(
    fpath=__file__, archive_commit="$Format:%h$", reponame="lumen"))
